/*
** Copyright (C) 2004 Jesse Chappell <jesse@essej.net>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/

#ifndef __sooperlooper_pixmap_includes__
#define __sooperlooper_pixmap_includes__

// include all the default pixmaps

#include "pixmaps/1x_rate_active.xpm"
#include "pixmaps/1x_rate_disabled.xpm"
#include "pixmaps/1x_rate_focus.xpm"
#include "pixmaps/1x_rate_normal.xpm"
#include "pixmaps/1x_rate_selected.xpm"
#include "pixmaps/delay_active.xpm"
#include "pixmaps/delay_disabled.xpm"
#include "pixmaps/delay_focus.xpm"
#include "pixmaps/delay_normal.xpm"
#include "pixmaps/delay_selected.xpm"
#include "pixmaps/double_rate_active.xpm"
#include "pixmaps/double_rate_disabled.xpm"
#include "pixmaps/double_rate_focus.xpm"
#include "pixmaps/double_rate_normal.xpm"
#include "pixmaps/double_rate_selected.xpm"
#include "pixmaps/half_rate_active.xpm"
#include "pixmaps/half_rate_disabled.xpm"
#include "pixmaps/half_rate_focus.xpm"
#include "pixmaps/half_rate_normal.xpm"
#include "pixmaps/half_rate_selected.xpm"
#include "pixmaps/insert_active.xpm"
#include "pixmaps/insert_disabled.xpm"
#include "pixmaps/insert_focus.xpm"
#include "pixmaps/insert_normal.xpm"
#include "pixmaps/insert_selected.xpm"
#include "pixmaps/load_active.xpm"
#include "pixmaps/load_disabled.xpm"
#include "pixmaps/load_focus.xpm"
#include "pixmaps/load_normal.xpm"
#include "pixmaps/load_selected.xpm"
#include "pixmaps/multiply_active.xpm"
#include "pixmaps/multiply_disabled.xpm"
#include "pixmaps/multiply_focus.xpm"
#include "pixmaps/multiply_normal.xpm"
#include "pixmaps/multiply_selected.xpm"
#include "pixmaps/mute_active.xpm"
#include "pixmaps/mute_disabled.xpm"
#include "pixmaps/mute_focus.xpm"
#include "pixmaps/mute_normal.xpm"
#include "pixmaps/mute_selected.xpm"
#include "pixmaps/once_active.xpm"
#include "pixmaps/once_disabled.xpm"
#include "pixmaps/once_focus.xpm"
#include "pixmaps/once_normal.xpm"
#include "pixmaps/once_selected.xpm"
#include "pixmaps/overdub_active.xpm"
#include "pixmaps/overdub_disabled.xpm"
#include "pixmaps/overdub_focus.xpm"
#include "pixmaps/overdub_normal.xpm"
#include "pixmaps/overdub_selected.xpm"
#include "pixmaps/record_active.xpm"
#include "pixmaps/record_disabled.xpm"
#include "pixmaps/record_focus.xpm"
#include "pixmaps/record_normal.xpm"
#include "pixmaps/record_selected.xpm"
#include "pixmaps/redo_active.xpm"
#include "pixmaps/redo_disabled.xpm"
#include "pixmaps/redo_focus.xpm"
#include "pixmaps/redo_normal.xpm"
#include "pixmaps/redo_selected.xpm"
#include "pixmaps/replace_active.xpm"
#include "pixmaps/replace_disabled.xpm"
#include "pixmaps/replace_focus.xpm"
#include "pixmaps/replace_normal.xpm"
#include "pixmaps/replace_selected.xpm"
#include "pixmaps/reverse_active.xpm"
#include "pixmaps/reverse_disabled.xpm"
#include "pixmaps/reverse_focus.xpm"
#include "pixmaps/reverse_normal.xpm"
#include "pixmaps/reverse_selected.xpm"
#include "pixmaps/save_active.xpm"
#include "pixmaps/save_disabled.xpm"
#include "pixmaps/save_focus.xpm"
#include "pixmaps/save_normal.xpm"
#include "pixmaps/save_selected.xpm"
#include "pixmaps/scratch_active.xpm"
#include "pixmaps/scratch_disabled.xpm"
#include "pixmaps/scratch_focus.xpm"
#include "pixmaps/scratch_normal.xpm"
#include "pixmaps/scratch_selected.xpm"
#include "pixmaps/trig_active.xpm"
#include "pixmaps/trig_disabled.xpm"
#include "pixmaps/trig_focus.xpm"
#include "pixmaps/trig_normal.xpm"
#include "pixmaps/trig_selected.xpm"
#include "pixmaps/undo_active.xpm"
#include "pixmaps/undo_disabled.xpm"
#include "pixmaps/undo_focus.xpm"
#include "pixmaps/undo_normal.xpm"
#include "pixmaps/undo_selected.xpm"
#include "pixmaps/substitute_active.xpm"
#include "pixmaps/substitute_disabled.xpm"
#include "pixmaps/substitute_focus.xpm"
#include "pixmaps/substitute_normal.xpm"
#include "pixmaps/substitute_selected.xpm"
#include "pixmaps/pause_active.xpm"
#include "pixmaps/pause_disabled.xpm"
#include "pixmaps/pause_focus.xpm"
#include "pixmaps/pause_normal.xpm"
#include "pixmaps/pause_selected.xpm"
#include "pixmaps/solo_active.xpm"
#include "pixmaps/solo_disabled.xpm"
#include "pixmaps/solo_focus.xpm"
#include "pixmaps/solo_normal.xpm"
#include "pixmaps/solo_selected.xpm"


#endif
